/************************************************************************************************
 * Copyright  (c) 2014, 2015 University of Pennsylvania
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose, without fee, and without a
 * written agreement is hereby granted, provided that the above copyright
 * notice and this paragraph and the following two paragraphs appear in
 * all copies.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF PENNSYLVANIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF PENNSYLVANIA
 * HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF PENNSYLVANIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON
 * AN "AS IS" BASIS, AND THE UNIVERSITY OF PENNSYLVANIA HAS NO OBLIGATIONS TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 ***********************************************************************************************/

/*** PowerOpt.java

Author: Edin Kadric
	IC Group, University of Pennsylvania

	Available online: http://ic.ese.upenn.edu/distributions/meme_fpga2015/

	Optimizes memory mapping for power reduction
	See our FPGA 2015 paper:
		Edin Kadric, David Lakata, Andre Dehon, "Impact of Memory Architecture on FPGA Energy Consumption"

 ***/

import java.io.*;
import java.util.ArrayList;

public class PowerOpt {
	public static void main(String[] args){
		if(args.length!=9){
			System.out.println("Usage: java PowerOpt <path> <benchmark> <phy_depth1> <phy_width> <mode> <delayopt> <force_low_depth> <phy_depth2> <phy_width2>");
		}else{
			//Arguments
			String path = args[0]; // path to benchmark
			String benchmark = args[1]; // benchmark name
			int physical_depth = Integer.parseInt(args[2]); // depth of physical memory
			int physical_width = Integer.parseInt(args[3]); // width of physical memory (mostly tested with 32)
			int mode = Integer.parseInt(args[4]); //Unused
			int delayopt = Integer.parseInt(args[5]); // set to 1 to avoid the PowerOpt mapping (typically will be 0)
			int force_lowdepth = Integer.parseInt(args[6]); //set to 0 in general, otherwise this can set manually the extent to which PowerOpt is applied
			int physical_depth2 = Integer.parseInt(args[7]); // depth of second physical memory; if only one present, then set to 0
			int physical_width2 = Integer.parseInt(args[8]); // depth of second physical memory; if only one present, then set to 0

			try{
				//Variables and files
				BufferedReader in0 = new BufferedReader(new FileReader(path + benchmark + ".v")); //input verilog file
				BufferedWriter out = new BufferedWriter(new FileWriter(path + benchmark + "Opt.v")); //output verilog file
				String s = in0.readLine();
				boolean limitStudy = false; //limit study --- Usually will be FALSE
				if(physical_depth==123456789){
					limitStudy = true;
				}
				int low_depth = physical_width; //logical width to use; defaults to physical width, will be updated later
				int low_depth2 = physical_width2; //logical width to use; defaults to physical width, will be updated later
				int N = 0; //total memories seen
				int physical_depth_saved = physical_depth; // For mem-2, we have a second memory size
				int physical_width_saved = physical_width; // For mem-2, we have a second memory size
				int low_depth_saved = low_depth; // For mem-2, we have a second memory size

				///////////// The memory modules will get broken down into sub-memories according to width (index i) and depth (indek k)
				///////////// Each sub-memory will get a unique name with i and k embedded in it

				///////////// First do a pass over the verilog file to read the meminit comments (OPTIONAL)

				//If your tools support memory initialization (which VTR doesn't by default),
				//You can write a line as follows in your verilog file for each memory:
				//      //NEW MEMINIT: .name=(foo), .filename=(bar);
				//	This means that the memory module called foo is to be initialized by the binary file bar
				//	PowerOpt will keep that information in the output verilog file, broken down appropriately among the mapped memories
				ArrayList<String> mem_init_all_data = new ArrayList<String>(); //name of module; populate from verilog file
				ArrayList<String> mem_init_file = new ArrayList<String>(); //name of meminit file; populate from verilog file
				ArrayList<Integer> kList = new ArrayList<Integer>(); //index of the sub-memory by depth; populate as break down memories
				ArrayList<Integer> iList = new ArrayList<Integer>(); //index of the sub-memory by width; populate as break down memories
				ArrayList<String> module0List_no_i = new ArrayList<String>(); //module name without i (one per group of sub-memories); populate as break down memories
				ArrayList<Integer> dList = new ArrayList<Integer>(); //depth of the sub-memory; populate as break down memories
				ArrayList<Integer> wList = new ArrayList<Integer>(); //width of the sub-memory; populate as break down memories
				ArrayList<Integer> wtotalList = new ArrayList<Integer>(); //total width of the group of sub-memories; populate as break down memories

				while(s!=null){
					if(s.indexOf("//NEW MEMINIT: ")==0){
						mem_init_all_data.add(isolate(s, ".name=(", ")"));
						System.out.println("Add to list of module names in PowerOpt.java: " + mem_init_all_data.get(mem_init_all_data.size()-1));
						mem_init_file.add(isolate(s, ".filename=(", ")"));
					}
					s = in0.readLine();
				}
				System.out.println("# of Memories that need to be initialized -> " + mem_init_all_data.size());
				System.out.println();
				in0.close();

				/////////////
				///////////// Now read through verilog file again and map the memories
				/////////////

				//re-open file
				BufferedReader in = new BufferedReader(new FileReader(path + benchmark + ".v"));
				s = in.readLine();

				while(s!=null){
					if(s.indexOf("dual_port_ram")>=0){ // FIRST CASE: DUAL-PORTED

						// add a memory (for unique node names)
						N++;

						//Read all memory info (could be broken down into multiple lines)
						while(!(s.indexOf(";")>=0)){
							s = s + in.readLine();
						}

						//Isolate net names
						String moduleName = isolate(s, "dual_port_ram", "(");
						String moduleName0 = moduleName.replace(" ", "");
						String clkName = isolate(s, ".clk(", ")");
						String addr1Name = isolate(s, ".addr1(", ")");
						String addr2Name = isolate(s, ".addr2(", ")");
						String data1Name = isolate(s, ".data1(", ")");
						String data2Name = isolate(s, ".data2(", ")");
						String we1Name = isolate(s, ".we1(", ")");
						String we2Name = isolate(s, ".we2(", ")");
						String out1Name = isolate(s, ".out1(", ")");
						String out2Name = isolate(s, ".out2(", ")");
						String enName = isolate(s, ".en(", ")");

						//Read width and depth
						int width = 0;
						int depth = 0;
						if((!(s.indexOf(".width(")>=0)) || (!(s.indexOf(".depth(")>=0))){
							System.out.println("Error in PowerOpt.java: Need to provide depth and width for memories");
						}else{
							width = Integer.parseInt(isolate(s, ".width(", ")"));
							depth = Integer.parseInt(isolate(s, ".depth(", ")"));
						}

						//Read dp mode
						int dpMode = 0; // 0: true dual port - 1: simple dual port with write on port1 - 2: simple dual port with write on port2
						if(!(s.indexOf(".dp(")>=0)){
							System.out.println("Error in PowerOpt.java: Need to provide dp_mode for dual ported memories");
						}else{
							String dp = isolate(s, ".dp(", ")");
							if(dp.equals("true"))
								dpMode = 0;
							else if(dp.equals("simple1"))
								dpMode = 1;
							else if(dp.equals("simple2"))
								dpMode = 2;
							else
								System.out.println("Error in PowerOpt.java: Unrecognized dp mode " + dp);
						}

						//if true dual port then max width is divided by 2
						if(dpMode==0)
							physical_width /= 2;

						//if simple2 mode, then invert ports
						if(dpMode==2){
							String swap = "";
							swap = addr1Name; addr1Name = addr2Name; addr2Name = swap;
							swap = data1Name; data1Name = data2Name; data2Name = swap;
							swap = we1Name; we1Name = we2Name; we2Name = swap;
							swap = out1Name; out1Name = out2Name; out2Name = swap;
						}

						//List of memories to be written - after fixing width
						ArrayList<String> moduleList = new ArrayList<String>();
						ArrayList<String> addr1List = new ArrayList<String>();
						ArrayList<String> addr2List = new ArrayList<String>();
						ArrayList<String> data1List = new ArrayList<String>();
						ArrayList<String> data2List = new ArrayList<String>();
						ArrayList<String> enList = new ArrayList<String>();
						ArrayList<String> out1List = new ArrayList<String>();
						ArrayList<String> out2List = new ArrayList<String>();
						ArrayList<String> signalsToAddList = new ArrayList<String>();
						ArrayList<Integer> widthList = new ArrayList<Integer>();
						ArrayList<Integer> depthList = new ArrayList<Integer>();

						//List of memories to be written - after fixing depth (after sizing)
						ArrayList<String> moduleListSized = new ArrayList<String>();
						ArrayList<String> addr1ListSized = new ArrayList<String>();
						ArrayList<String> addr2ListSized = new ArrayList<String>();
						ArrayList<String> data1ListSized = new ArrayList<String>();
						ArrayList<String> data2ListSized = new ArrayList<String>();
						ArrayList<String> enListSized = new ArrayList<String>();
						ArrayList<String> weListSized = new ArrayList<String>();
						ArrayList<String> out1ListSized = new ArrayList<String>();
						ArrayList<String> out2ListSized = new ArrayList<String>();
						ArrayList<String> signalsToAddListSized = new ArrayList<String>();

						// Mem-2 piece
						// Estimate whether the given memory should be mapped to the smaller memory or the larger one
						boolean mapToLargeMemory = false;
						if(!limitStudy){
							if((depth*width>=(physical_depth2*physical_width2/2)) && (physical_depth2!=0)){
								mapToLargeMemory = true;
								physical_depth = physical_depth2; //swap them
								physical_width = physical_width2; //swap them
								low_depth = low_depth2; //swap them
							}
						}

						//Estimate optimal low_depth
						int memories_needed = (depth/physical_depth);
						int widthMaxed = (nextPowerOf2(width)<=physical_width)?nextPowerOf2(width):physical_width;
						memories_needed /= (physical_width / widthMaxed);
						if(delayopt==0){
							if((memories_needed<=4) || (widthMaxed<=1)){
								low_depth = widthMaxed; //activate 1
							}else if((memories_needed<=8) || (widthMaxed<=2)){
								low_depth = widthMaxed/2; //activate 2
							}else if((memories_needed<=16) || (widthMaxed<=4)){
								low_depth = widthMaxed/4; //activate 4
							}else if((memories_needed<=32) || (widthMaxed<=8)){
								low_depth = widthMaxed/8; //activate 8
							}else if((memories_needed<=64) || (widthMaxed<=16)){
								low_depth = widthMaxed/16; //activate 16
							}else{
								low_depth = widthMaxed/32; //activate 32
							}
						}else{
							low_depth = 1; //if delayopt mode; them go down to width of 1 before we start playing the power-opt trick
						}
						if(force_lowdepth!=0){
							low_depth = force_lowdepth; //optionally we can force this to a particular value
						}

						//Break down the width
						int widthLeft = width;
						wtotalList.add(widthLeft);
						int i;
						for(i=0;widthLeft>0;i++){
							//Find next width
							int nextWidth = 0;
							int zeroBitsWidth = 0;
							int physical_width_mode = physical_width;
							while(physical_width_mode>=low_depth){
								if(widthLeft-physical_width_mode<0){
									nextWidth = widthLeft;
									zeroBitsWidth = nextPowerOf2(nextWidth)-nextWidth;
								}else{
									nextWidth = physical_width_mode;
									zeroBitsWidth = 0;
								}
								//Check capacity and see if could bring it down more
								int depthPhysicalMode = physical_depth * (physical_width / physical_width_mode);
								if((depth<=depthPhysicalMode) || limitStudy)
									break;
								//update for next try
								physical_width_mode /= 2;
							}
							//add info
							widthList.add(nextWidth);
							wList.add(nextWidth);
							depthList.add(depth);
							moduleList.add(moduleName + "_" + i);
							iList.add(i); //for mem init
							addr1List.add(addr1Name);
							addr2List.add(addr2Name);
							enList.add(enName);
							if(zeroBitsWidth==0){
								out1List.add("mem_out1_power_opt_break_width_" + i + "_n" + N);
								out2List.add("mem_out2_power_opt_break_width_" + i + "_n" + N);
								String toAdd = "";
								if(dpMode==0){
									toAdd += "wire [" + (nextWidth-1) + ":0] mem_out1_power_opt_break_width_" + i + "_n" + N + ";\n"
											+ "assign " + out1Name + "[" + (widthLeft-1) + ":" + (widthLeft-nextWidth) + "] = mem_out1_power_opt_break_width_" + i + "_n" + N + "[" + (nextWidth-1) + ":0];\n";
								}						
								toAdd += "wire [" + (nextWidth-1) + ":0] mem_out2_power_opt_break_width_" + i + "_n" + N + ";\n"
										+ "assign " + out2Name + "[" + (widthLeft-1) + ":" + (widthLeft-nextWidth) + "] = mem_out2_power_opt_break_width_" + i + "_n" + N + "[" + (nextWidth-1) + ":0];\n";
								signalsToAddList.add(toAdd);
								data1List.add(data1Name + "[" + (widthLeft-1) + ":" + (widthLeft-nextWidth) + "]");
								data2List.add(data2Name + "[" + (widthLeft-1) + ":" + (widthLeft-nextWidth) + "]");
							}else{
								String zeroString = "" + zeroBitsWidth + "'b0";
								out1List.add("mem_out1_power_opt_break_width_" + i + "_n" + N);
								out2List.add("mem_out2_power_opt_break_width_" + i + "_n" + N);
								String toAdd = "";
								if(dpMode==0){
									toAdd += "wire [" + (nextPowerOf2(nextWidth)-1) + ":0] mem_out1_power_opt_break_width_" + i + "_n" + N + ";\n"
											+ "assign " + out1Name + "[" + (widthLeft-1) + ":" + (widthLeft-nextWidth) + "] = mem_out1_power_opt_break_width_" + i + "_n" + N + "[" + (widthLeft-1) + ":" + (widthLeft-nextWidth) + "];\n";
								}
								toAdd += "wire [" + (nextPowerOf2(nextWidth)-1) + ":0] mem_out2_power_opt_break_width_" + i + "_n" + N + ";\n"
										+ "assign " + out2Name + "[" + (widthLeft-1) + ":" + (widthLeft-nextWidth) + "] = mem_out2_power_opt_break_width_" + i + "_n" + N + "[" + (widthLeft-1) + ":" + (widthLeft-nextWidth) + "];\n";
								signalsToAddList.add(toAdd);
								data1List.add("{" + zeroString + ", " + data1Name + "[" + (widthLeft-1) + ":" + (widthLeft-nextWidth) + "]}");
								data2List.add("{" + zeroString + ", " + data2Name + "[" + (widthLeft-1) + ":" + (widthLeft-nextWidth) + "]}");
							}
							widthLeft -= nextWidth;
						}
						module0List_no_i.add(moduleName0); //for mem init - add only once per i

						//Write signals and assignments
						out.write("\n//wires and assignments\n");
						for(int j=0;j<moduleList.size();j++){
							String toWrite = "";
							toWrite += (signalsToAddList.get(j));
							out.write(toWrite);
						}

						//Once width is broken down, make sure #bits fits into (width x depth); if not, then need more memories to implement this
						for(int j=0;j<i;j++){
							width = widthList.get(j);
							depth = depthList.get(j);
							int bitsNeeded = width * depth;
							int capacity = physical_width * physical_depth;
							//normal case, don't do anything
							if((bitsNeeded<=capacity) || limitStudy){
								moduleListSized.add(moduleList.get(j));
								addr1ListSized.add(addr1List.get(j));
								addr2ListSized.add(addr2List.get(j));
								data1ListSized.add(data1List.get(j));
								data2ListSized.add(data2List.get(j));
								enListSized.add(enList.get(j));
								weListSized.add(we1Name);
								out1ListSized.add(out1List.get(j));
								out2ListSized.add(out2List.get(j));
								signalsToAddListSized.add("");
								//if no k added:
								kList.add(-1); //for mem init
								dList.add(depth); //for mem init
							}else{
								//Find the number of memories to instantiate; decrease depth; don't pack onto same width
								//this may decrease the number of memories but may also activate more of them and increase routing
								//Note: assumption is that depth is always a power-of-2
								int depthLeft = depth;
								int depthModeUsed = capacity / nextPowerOf2(width);
								int k;
								for(k=0;depthLeft>0;k++){
									int nextDepth = depthModeUsed;
									//Next memory has depth: nextDepth, and width: width
									moduleListSized.add(moduleList.get(j) + "_" + k);
									kList.add(k); //for mem init
									int actualDepth = (nextDepth<depthLeft)?nextDepth:depthLeft; //min(depthLeft, nextDepth) -> gives actual depth used by memory at given k
									dList.add(actualDepth); //for mem init
									addr1ListSized.add(addr1List.get(j) + "_id" + j + "_n" + N + "_mux" + k);
									addr2ListSized.add(addr2List.get(j) + "_id" + j + "_n" + N + "_mux" + k);
									enListSized.add("topbits_" + j + "_n" + N + "_mux" + k);
									weListSized.add("we_sig_" + j + "_n" + N + "_mux" + k);
									signalsToAddListSized.add("");
									out1ListSized.add(format(out1List.get(j)) + "_mux" + k);
									out2ListSized.add(format(out2List.get(j)) + "_mux" + k);
									data1ListSized.add(format(data1List.get(j)) + "_mux" + k);
									data2ListSized.add(format(data2List.get(j)) + "_mux" + k);

									//decrement depthLeft
									depthLeft -= nextDepth;
								}

								//add control logic to select 'data' and 'out' using top bits of address
								//a little different from single port case because here could enable two memories at once
								//note: sort combinational loop bug showed: need to register select bits for addr2!!
								int selectBitCount = log(k);
								int addrBitCount = log(depth)-selectBitCount;
								String addrString = "";
								//select signal for 'en' and 'out' and 'data'
								String caseString = "";
								if(selectBitCount!=1){
									caseString += "wire [" + (selectBitCount-1) + ":0] " + "topbitsEN1_" + j + "_n" + N + ";\n";
									caseString += "wire [" + (selectBitCount-1) + ":0] " + "topbitsEN2_" + j + "_n" + N + ";\n";
								}else{
									caseString += "wire " + "topbitsEN1_" + j + "_n" + N + ";\n";
									caseString += "wire " + "topbitsEN2_" + j + "_n" + N + ";\n";
								}
								if(selectBitCount!=1){
									caseString += "reg [" + (selectBitCount-1) + ":0] " + "reg_topbitsEN1_" + j + "_n" + N + ";\n";
									caseString += "reg [" + (selectBitCount-1) + ":0] " + "reg_topbitsEN2_" + j + "_n" + N + ";\n";
								}else{
									caseString += "reg " + "reg_topbitsEN1_" + j + "_n" + N + ";\n";
									caseString += "reg " + "reg_topbitsEN2_" + j + "_n" + N + ";\n";
								}
								caseString += "assign " + "topbitsEN1_" + j + "_n" + N + " = " + addr1List.get(j) + "[" + (addrBitCount+selectBitCount-1) + ":" + (addrBitCount) + "];\n";
								caseString += "assign " + "topbitsEN2_" + j + "_n" + N + " = " + addr2List.get(j) + "[" + (addrBitCount+selectBitCount-1) + ":" + (addrBitCount) + "];\n";
								caseString += "always@(posedge " + clkName + ") begin " + "reg_topbitsEN1_" + j + "_n" + N + " <= " + addr1List.get(j) + "[" + (addrBitCount+selectBitCount-1) + ":" + (addrBitCount) + "]; end\n";
								caseString += "always@(posedge " + clkName + ") begin " + "reg_topbitsEN2_" + j + "_n" + N + " <= " + addr2List.get(j) + "[" + (addrBitCount+selectBitCount-1) + ":" + (addrBitCount) + "]; end\n";
								for(int m=0;m<k;m++){
									String currentSelect1 = "(topbitsEN1_" + j + "_n" + N + "==" + selectBitCount + "'d" + m + ")";
									String currentSelect2 = "(topbitsEN2_" + j + "_n" + N + "==" + selectBitCount + "'d" + m + ")";
									String currentSelectAddr1 = (false)?currentSelect1:"(1==1)"; //if true then set addr to 0 when memory not used
									String currentSelectAddr2 = (false)?currentSelect2:"(1==1)"; //if true then set addr to 0 when memory not used
									//addr that is sent to module
									addrString += "wire [" + (addrBitCount-1) + ":0] " + addr1List.get(j) + "_id" + j + "_n" + N + "_mux" + m + ";\n";
									addrString += "wire [" + (addrBitCount-1) + ":0] " + addr2List.get(j) + "_id" + j + "_n" + N + "_mux" + m + ";\n";
									addrString += "assign " + addr1List.get(j) + "_id" + j + "_n" + N + "_mux" + m + " = " + currentSelectAddr1 + "?(" + addr1List.get(j) + "[" + (addrBitCount-1) + ":0]):0; //" + (addr1List.get(j) + "_id" + j + "_n" + N + "_mux" + m) + "[" + (addrBitCount-1) + ":0];\n";
									addrString += "assign " + addr2List.get(j) + "_id" + j + "_n" + N + "_mux" + m + " = " + currentSelectAddr2 + "?(" + addr2List.get(j) + "[" + (addrBitCount-1) + ":0]):0; //" + (addr2List.get(j) + "_id" + j + "_n" + N + "_mux" + m) + "[" + (addrBitCount-1) + ":0];\n";
									//case statement for 'en' and 'out' and 'data' (turns out it's not a case statement, just a bunch of conditions)
									//en
									caseString += "wire topbits1_" + j + "_n" + N + "_mux" + m + ";\n";
									caseString += "wire topbits2_" + j + "_n" + N + "_mux" + m + ";\n";
									caseString += "assign topbits1_" + j + "_n" + N + "_mux" + m + " = " + currentSelect1 + " ? " + enList.get(j) + " : 1'b0;\n";
									caseString += "assign topbits2_" + j + "_n" + N + "_mux" + m + " = " + currentSelect2 + " ? " + enList.get(j) + " : 1'b0;\n";
									//combine above two into overall enable for memory
									caseString += "wire topbits_" + j + "_n" + N + "_mux" + m + ";\n";
									caseString += "assign topbits_" + j + "_n" + N + "_mux" + m + " = " + "topbits1_" + j + "_n" + N + "_mux" + m + " | topbits2_" + j + "_n" + N + "_mux" + m + ";\n";
									//write-enable - only writing to one of the banks
									caseString += "wire we_sig_" + j + "_n" + N + "_mux" + m + ";\n";
									caseString += "assign we_sig_" + j + "_n" + N + "_mux" + m + " = " + currentSelect1 + " ? " + we1Name + " : 1'b0;\n";
									//data
									caseString += "wire [" + (width-1) + ":0] " + format(data1List.get(j)) + "_mux" + m + ";\n";
									caseString += "assign " + format(data1List.get(j)) + "_mux" + m + " = " + data1List.get(j) + ";\n";
									if(dpMode==0){
										caseString += "wire [" + (width-1) + ":0] " + format(data2List.get(j)) + "_mux" + m + ";\n";
										caseString += "assign " + format(data2List.get(j)) + "_mux" + m + " = " + data2List.get(j) + ";\n";
									}
									//out (just declare)
									if(dpMode==0)
										caseString += "wire [" + (width-1) + ":0] " + format(out1List.get(j)) + "_mux" + m + ";\n";
									caseString += "wire [" + (width-1) + ":0] " + format(out2List.get(j)) + "_mux" + m + ";\n";

								}
								//out (now assign)
								//FIRST PORT
								if(dpMode==0){
									caseString += "assign " + out1List.get(j) + " = \n";
									for(int m=0;m<k;m++){
										caseString += "(reg_topbitsEN1_" + j + "_n" + N + "==" + selectBitCount + "'d" + m + ") ? " + format(out1List.get(j)) + "_mux" + m + ": (\n";
									}
									caseString += "0)";
									for(int m=0;m<k-1;m++){
										caseString += ")";
									}
									caseString += ";\n";
								}
								//SECOND PORT
								caseString += "assign " + out2List.get(j) + " = \n";
								for(int m=0;m<k;m++){
									caseString += "(reg_topbitsEN2_" + j + "_n" + N + "==" + selectBitCount + "'d" + m + ") ? " + format(out2List.get(j)) + "_mux" + m + ": (\n";
								}
								caseString += "0)";
								for(int m=0;m<k-1;m++){
									caseString += ")";
								}
								caseString += ";\n";

								//update signals to add
								String addThis = caseString + addrString;
								signalsToAddListSized.set(j, addThis);
							}
						}

						//write mapped memories to output verilog file
						for(int j=0;j<moduleListSized.size();j++){
							String toWrite = "";
							toWrite += ("\n");
							toWrite += ("//Start a new memory\n");
							toWrite += (signalsToAddListSized.get(j));
							if(dpMode==0)
								toWrite += ("dual_port_ram_poweropt" + (mapToLargeMemory?"2":"") + moduleListSized.get(j) + "(\n");
							else
								toWrite += ("dual_port_ram_poweropt_s" + (mapToLargeMemory?"2":"") + moduleListSized.get(j) + "(\n");
							toWrite += (".clk(" + clkName + "),\n");
							toWrite += (".addr1(" + addr1ListSized.get(j) + "),\n");
							toWrite += (".addr2(" + addr2ListSized.get(j) + "),\n");
							toWrite += (".data1(" + data1ListSized.get(j) + "),\n");
							if(dpMode==0)
								toWrite += (".data2(" + data2ListSized.get(j) + "),\n");
							toWrite += (".writeen1(" + weListSized.get(j) + "),\n");
							if(dpMode==0)
								toWrite += (".writeen2(" + we2Name + "),\n");
							if(dpMode==0)
								toWrite += (".readen1(1'b1),\n");
							toWrite += (".readen2(1'b1),\n");
							if(delayopt==1) //force all on for delayopt
								toWrite += (".clken(1'b1),\n");
							else //standard for poweropt:
								toWrite += (".clken(" + enListSized.get(j) + "),\n");
							if(dpMode==0)
								toWrite += (".out1(" + out1ListSized.get(j) + "),\n");
							toWrite += (".out2(" + out2ListSized.get(j) + ")\n");
							toWrite += (");\n");
							out.write(toWrite);
						}

						//if true dual port then max width is divided by 2 - here bring it back to normal
						if(dpMode==0)
							physical_width *= 2;

						//Mem-2, restore normal state after this large memory mapping
						if(!limitStudy){
							if(mapToLargeMemory){
								mapToLargeMemory = false;
								physical_depth = physical_depth_saved; //swap them back
								physical_width = physical_width_saved; //swap them back
								low_depth = low_depth_saved; //swap them back
							}
						}

					}else if(s.indexOf("single_port_ram")>=0){ // SECOND CASE: SINGLE-PORTED (very similar to first case above)

						// add a memory (for unique node names)
						N++;

						//Read all memory info (could be broken down into multiple lines)
						while(!(s.indexOf(";")>=0)){
							s = s + in.readLine();
						}

						//Isolate net names
						String moduleName = isolate(s, "single_port_ram", "(");
						String moduleName0 = moduleName.replace(" ", "");
						String clkName = isolate(s, ".clk(", ")");
						String addrName = isolate(s, ".addr(", ")");
						String dataName = isolate(s, ".data(", ")");
						String weName = isolate(s, ".we(", ")");
						String outName = isolate(s, ".out(", ")");
						String enName = isolate(s, ".en(", ")");

						//Read width and depth
						int width = 0;
						int depth = 0;
						if((!(s.indexOf(".width(")>=0)) || (!(s.indexOf(".depth(")>=0))){
							System.out.println("Error in PowerOpt.java: Need to provide depth and width for memories");
						}else{
							width = Integer.parseInt(isolate(s, ".width(", ")"));
							depth = Integer.parseInt(isolate(s, ".depth(", ")"));
						}

						//List of memories to be written - after fixing width
						ArrayList<String> moduleList = new ArrayList<String>();
						ArrayList<String> addrList = new ArrayList<String>();
						ArrayList<String> dataList = new ArrayList<String>();
						ArrayList<String> enList = new ArrayList<String>();
						ArrayList<String> outList = new ArrayList<String>();
						ArrayList<String> signalsToAddList = new ArrayList<String>();
						ArrayList<Integer> widthList = new ArrayList<Integer>();
						ArrayList<Integer> depthList = new ArrayList<Integer>();

						//List of memories to be written - after fixing depth (after sizing)
						ArrayList<String> moduleListSized = new ArrayList<String>();
						ArrayList<String> addrListSized = new ArrayList<String>();
						ArrayList<String> dataListSized = new ArrayList<String>();
						ArrayList<String> enListSized = new ArrayList<String>();
						ArrayList<String> weListSized = new ArrayList<String>();
						ArrayList<String> outListSized = new ArrayList<String>();
						ArrayList<String> signalsToAddListSized = new ArrayList<String>();

						// Mem-2 piece
						// Estimate whether the given memory should be mapped to the smaller memory or the larger one
						boolean mapToLargeMemory = false;
						if(!limitStudy){
							if((depth*width>=(physical_depth2*physical_width2/2)) && (physical_depth2!=0)){
								mapToLargeMemory = true;
								physical_depth = physical_depth2; //swap them
								physical_width = physical_width2; //swap them
								low_depth = low_depth2; //swap them
							}
						}

						//Estimate optimal low_depth
						int memories_needed = (depth/physical_depth);
						int widthMaxed = (nextPowerOf2(width)<=physical_width)?nextPowerOf2(width):physical_width;
						memories_needed /= (physical_width / widthMaxed);
						if(delayopt==0){
							if((memories_needed<=4) || (widthMaxed<=1)){
								low_depth = widthMaxed; //activate 1
							}else if((memories_needed<=8) || (widthMaxed<=2)){
								low_depth = widthMaxed/2; //activate 2
							}else if((memories_needed<=16) || (widthMaxed<=4)){
								low_depth = widthMaxed/4; //activate 4
							}else if((memories_needed<=32) || (widthMaxed<=8)){
								low_depth = widthMaxed/8; //activate 8
							}else if((memories_needed<=64) || (widthMaxed<=16)){
								low_depth = widthMaxed/16; //activate 16
							}else{
								low_depth = widthMaxed/32; //activate 32
							}
						}else{
							low_depth = 1; //if delayopt mode; them go down to width of 1 before we start playing the power-opt trick
						}
						if(force_lowdepth!=0){
							low_depth = force_lowdepth; //optionally we can force this to a particular value
						}

						//Break down the width
						int widthLeft = width;
						wtotalList.add(widthLeft);
						int i;
						for(i=0;widthLeft>0;i++){
							//Find next width
							int nextWidth = 0;
							int zeroBitsWidth = 0;
							int physical_width_mode = physical_width;
							while(physical_width_mode>=low_depth){
								if(widthLeft-physical_width_mode<0){
									nextWidth = widthLeft;
									zeroBitsWidth = nextPowerOf2(nextWidth)-nextWidth;
								}else{
									nextWidth = physical_width_mode;
									zeroBitsWidth = 0;
								}
								//Check capacity and see if could bring it down more
								int depthPhysicalMode = physical_depth * (physical_width / physical_width_mode);
								if((depth<=depthPhysicalMode) || limitStudy)
									break;
								//update for next try
								physical_width_mode /= 2;
							}
							//add info
							widthList.add(nextWidth);
							wList.add(nextWidth);
							depthList.add(depth);
							moduleList.add(moduleName + "_" + i);
							iList.add(i); //for mem init
							addrList.add(addrName);
							enList.add(enName);
							if(zeroBitsWidth==0){
								outList.add("mem_out_power_opt_break_width_" + i + "_n" + N);
								signalsToAddList.add("wire [" + (nextWidth-1) + ":0] mem_out_power_opt_break_width_" + i + "_n" + N + ";\n"
										+ "assign " + outName + "[" + (widthLeft-1) + ":" + (widthLeft-nextWidth) + "] = mem_out_power_opt_break_width_" + i + "_n" + N + "[" + (nextWidth-1) + ":0];\n");
								dataList.add(dataName + "[" + (widthLeft-1) + ":" + (widthLeft-nextWidth) + "]");
							}else{
								String zeroString = "" + zeroBitsWidth + "'b0";
								outList.add("mem_out_power_opt_break_width_" + i + "_n" + N);
								signalsToAddList.add("wire [" + (nextPowerOf2(nextWidth)-1) + ":0] mem_out_power_opt_break_width_" + i + "_n" + N + ";\n"
										+ "assign " + outName + "[" + (widthLeft-1) + ":" + (widthLeft-nextWidth) + "] = mem_out_power_opt_break_width_" + i + "_n" + N + "[" + (widthLeft-1) + ":" + (widthLeft-nextWidth) + "];\n");
								dataList.add("{" + zeroString + ", " + dataName + "[" + (widthLeft-1) + ":" + (widthLeft-nextWidth) + "]}");
							}
							widthLeft -= nextWidth;
						}
						module0List_no_i.add(moduleName0); //for mem init - add only once per i

						//Write signals and assignments
						out.write("\n//wires and assignments\n");
						for(int j=0;j<moduleList.size();j++){
							String toWrite = "";
							toWrite += (signalsToAddList.get(j));
							out.write(toWrite);
						}

						//Once width is broken down, make sure #bits fits into (width x depth); if not, then need more memories to implement this
						System.out.println("*New sp memory*");
						for(int j=0;j<i;j++){
							width = widthList.get(j);
							depth = depthList.get(j);
							int bitsNeeded = width * depth;
							int capacity = physical_width * physical_depth;
							//normal case, don't do anything
							if((bitsNeeded<=capacity) || limitStudy){
								moduleListSized.add(moduleList.get(j));
								addrListSized.add(addrList.get(j));
								dataListSized.add(dataList.get(j));
								enListSized.add(enList.get(j));
								weListSized.add(weName);
								outListSized.add(outList.get(j));
								signalsToAddListSized.add("");
								//if no k added:
								kList.add(-1); //for mem init
								dList.add(depth); //for mem init
							}else{
								//Find the number of memories to instantiate; decrease depth; don't pack onto same width
								//this may decrease the number of memories but may also activate more of them and increase routing
								//Note: assumption is that depth is always a power-of-2
								int depthLeft = depth;
								int depthModeUsed = capacity / nextPowerOf2(width);
								int k;
								for(k=0;depthLeft>0;k++){
									int nextDepth = depthModeUsed;
									//Next memory has depth: nextDepth, and width: width
									moduleListSized.add(moduleList.get(j) + "_" + k);
									kList.add(k); //for mem init
									int actualDepth = (nextDepth<depthLeft)?nextDepth:depthLeft; //min(depthLeft, nextDepth) -> gives actual depth used by memory at given k
									dList.add(actualDepth); //for mem init
									addrListSized.add(addrList.get(j) + "_id" + j + "_n" + N + "_mux" + k);
									enListSized.add("topbits_" + j + "_n" + N + "_mux" + k);
									weListSized.add("we_sig_" + j + "_n" + N + "_mux" + k);
									signalsToAddListSized.add("");
									outListSized.add(format(outList.get(j)) + "_mux" + k);
									dataListSized.add(format(dataList.get(j)) + "_mux" + k);

									//decrement depthLeft
									depthLeft -= nextDepth;
								}

								//add control logic to select 'data' and 'out' using top bits of address
								int selectBitCount = log(k);
								int addrBitCount = log(depth)-selectBitCount;
								String addrString = "";
								//select signal for 'en' and 'out' and 'data'
								String caseString = "";
								if(selectBitCount!=1){
									caseString += "wire [" + (selectBitCount-1) + ":0] " + "topbitsEN_" + j + "_n" + N + ";\n";
								}else{
									caseString += "wire " + "topbitsEN_" + j + "_n" + N + ";\n";
								}
								if(selectBitCount!=1){
									caseString += "reg [" + (selectBitCount-1) + ":0] " + "reg_topbitsEN_" + j + "_n" + N + ";\n";
								}else{
									caseString += "reg " + "reg_topbitsEN_" + j + "_n" + N + ";\n";
								}
								caseString += "assign " + "topbitsEN_" + j + "_n" + N + " = " + addrList.get(j) + "[" + (addrBitCount+selectBitCount-1) + ":" + (addrBitCount) + "];\n";
								caseString += "always@(posedge " + clkName + ") begin " + "reg_topbitsEN_" + j + "_n" + N + " <= " + addrList.get(j) + "[" + (addrBitCount+selectBitCount-1) + ":" + (addrBitCount) + "]; end\n";
								for(int m=0;m<k;m++){
									String currentSelect = "(topbitsEN_" + j + "_n" + N + "==" + selectBitCount + "'d" + m + ")";
									String currentSelectAddr = (false)?currentSelect:"(1==1)"; //if true then set addr to 0 when memory not used
									//addr that is sent to module
									addrString += "wire [" + (addrBitCount-1) + ":0] " + addrList.get(j) + "_id" + j + "_n" + N + "_mux" + m + ";\n";
									addrString += "assign " + addrList.get(j) + "_id" + j + "_n" + N + "_mux" + m + " = " + currentSelectAddr + "?(" + addrList.get(j) + "[" + (addrBitCount-1) + ":0]):0; //(" + addrList.get(j) + "_id" + j + "_n" + N + "_mux" + m + "[" + (addrBitCount-1) + ":0]);\n";
									//case statement for 'en' and 'out' and 'data' (turns out it's not a case statement, just a bunch of conditions)
									//en
									caseString += "wire topbits_" + j + "_n" + N + "_mux" + m + ";\n";
									caseString += "assign topbits_" + j + "_n" + N + "_mux" + m + " = " + currentSelect + " ? " + enList.get(j) + " : 1'b0;\n";
									//write-enable - only writing to one of the banks
									caseString += "wire we_sig_" + j + "_n" + N + "_mux" + m + ";\n";
									caseString += "assign we_sig_" + j + "_n" + N + "_mux" + m + " = " + currentSelect + " ? " + weName + " : 1'b0;\n";
									//data
									caseString += "wire [" + (width-1) + ":0] " + format(dataList.get(j)) + "_mux" + m + ";\n";
									caseString += "assign " + format(dataList.get(j)) + "_mux" + m + " = " + dataList.get(j) + ";\n";
									//out (just declare)
									caseString += "wire [" + (width-1) + ":0] " + format(outList.get(j)) + "_mux" + m + ";\n";

								}
								//out (now assign)
								caseString += "assign " + outList.get(j) + " = \n";
								for(int m=0;m<k;m++){
									caseString += "(reg_topbitsEN_" + j + "_n" + N + "==" + selectBitCount + "'d" + m + ") ? " + format(outList.get(j)) + "_mux" + m + ": (\n";
								}
								caseString += "0)";
								for(int m=0;m<k-1;m++){
									caseString += ")";
								}
								caseString += ";\n";

								//update signals to add
								String addThis = caseString + addrString;
								signalsToAddListSized.set(j, addThis);
							}
						}

						//write mapped memories to output verilog file
						for(int j=0;j<moduleListSized.size();j++){
							String toWrite = "";
							toWrite += ("\n");
							toWrite += ("//Start a new memory\n");
							toWrite += (signalsToAddListSized.get(j));
							toWrite += ("single_port_ram_poweropt" + (mapToLargeMemory?"2":"") + moduleListSized.get(j) + "(\n");
							toWrite += (".clk(" + clkName + "),\n");
							toWrite += (".addr(" + addrListSized.get(j) + "),\n");
							toWrite += (".data(" + dataListSized.get(j) + "),\n");
							toWrite += (".writeen(" + weListSized.get(j) + "),\n");
							toWrite += (".readen(1'b1),\n");
							if(delayopt==1) //force all on for delayopt
								toWrite += (".clken(1'b1),\n");
							else //standard for poweropt:
								toWrite += (".clken(" + enListSized.get(j) + "),\n");
							toWrite += (".out(" + outListSized.get(j) + ")\n");
							toWrite += (");\n");
							out.write(toWrite);
						}

						//Mem-2, restore normal state after this large memory mapping
						if(!limitStudy){
							if(mapToLargeMemory){
								mapToLargeMemory = false;
								physical_depth = physical_depth_saved; //swap them back
								physical_width = physical_width_saved; //swap them back
								low_depth = low_depth_saved; //swap them back
							}
						}

					}else{
						out.write(s + "\n");
					}

					s = in.readLine();
				}
				in.close();
				out.close();

				///////// 
				///////// Done, now use populated arraylists to set up info.meminit file (OPTIONAL)
				///////// 

				BufferedWriter infomeminit = new BufferedWriter(new FileWriter(path + "info.meminit"));
				int iCount = 0; //width divisions
				int kCount = 0; //depth divisions (-1 if depth untouched)
				for(int m=0;m<module0List_no_i.size();m++){
					//careful here: need count and update iCount and kCount even when no idx match!
					int idx = find(mem_init_all_data, module0List_no_i.get(m));
					String filename = "";
					if(idx!=-1){
						filename = mem_init_file.get(idx);
						//System.out.println("PowerOpt Debug: found match with " + module0List_no_i.get(m));
					}else{
						//System.out.println("PowerOpt Debug: didn't find match with " + module0List_no_i.get(m));
					}
					boolean firstLoop = true;
					int widthSum = wtotalList.get(m); //this will be decremented in the loop below; the MSBs of the width are given to the modules with lower index
					for(int i=0;;i++){
						if(iCount+i==iList.size()){ //if full
							iCount += i;
							break;
						} 
						if(iList.get(iCount+i)==0 && !firstLoop){ //if reached next set
							iCount += i;
							break;
						} 
						firstLoop = false;
						boolean firstLoop2 = true;
						int depthSum = 0;
						for(int k=0;;k++){
							if(kCount+k==kList.size()){ //if full
								kCount += k;
								break;
							}
							if(kList.get(kCount+k)<=0 && !firstLoop2){ //if reached next set
								kCount += k;
								break;
							}
							firstLoop2 = false;
							String newName = "";
							if(idx!=-1)
								newName = mem_init_all_data.get(idx) + "_" + iList.get(iCount+i);
							if(kList.get(kCount+k)!=-1){
								newName = newName + "_" + kList.get(kCount+k);
							}
							if(idx!=-1){
								infomeminit.write(".name(" + newName + ") .file(" + filename + ") " +
										".d0(" + (depthSum) + ") .d1(" + (depthSum+dList.get(kCount+k)) + ") " +
										".w0(" + (widthSum-wList.get(iCount+i)) + ") .w1(" + (widthSum) + ") \n");
							}
							depthSum += dList.get(kCount+k);
						}
						widthSum -= wList.get(iCount+i);
					}
				}
				infomeminit.close();

			}catch(Exception E){
				System.out.println("Exception in PowerOpt.java:");
				System.out.println(E);
			}
		}
	}

	//////////////////////////////////
	//////////// Functions ///////////
	//////////////////////////////////

	public static String isolate(String s, String a, String b){
		return s.substring(s.indexOf(a)+a.length(), (s.indexOf(a)+a.length())+s.substring(s.indexOf(a)+a.length()).indexOf(b));
	}

	public static String format(String s){
		s = s.replaceAll("\\[", "b_");
		s = s.replaceAll("\\]", "_b");
		s = s.replaceAll(":", "_");
		s = s.replaceAll("\\{", "b_");
		s = s.replaceAll("\\}", "_b");
		s = s.replaceAll(" ", "_");
		s = s.replaceAll(",", "_");
		s = s.replaceAll("'", "_");
		return s;
	}

	public static int nextPowerOf2(int in){
		int pow2 = 1;
		while(pow2<in){
			pow2 *= 2;
		}
		return pow2;
	}

	public static int log(int in){
		int log_in = 0;
		in /= 2;
		while(in!=0){
			in /= 2;
			log_in++;
		}
		return log_in;
	}

	public static int find(ArrayList<String> a, String element){
		for(int i=0;i<a.size();i++){
			if(a.get(i).equals(element))
				return i;
		}
		return -1;
	}
}
