######################
##### arch_gen.py ####
######################

# Based on VTR's architecture generation engine
# Modified by David Lakata and Edin Kadric

import sys
import re
import os
import math

openlist = []

##########################
##### Tech parameters ####
##########################

do_power = True

# Architecture Properties
Fc_in = 0.15
Fc_out = 0.1

power_short_circuit_percentage = 0.0

### the following parameters are used for delay calculation, they are the default VTR values, for 45nm, some of the values are scaled for technology (some are not???)
R_minW_nmos = 8926
R_minW_pmos = 16067
ipin_mux_trans_size = 1.222260

C_ipin_cblock = 1.47e-15
T_ipin_cblock = 7.247e-11

grid_logic_tile_area = 53894 #fix this soon

switch_R = 551
switch_Cin = 0.77e-15
switch_Cout = 4e-15
switch_T_del = 58e-12
switch_mux_trans_size = 2.630740
switch_buf_size = 27.645901

segment_Rmetal = 101
segment_Cmetal = 22.5e-15 #should not be used for power, change code to use Cwire instead

delay_crossbar_CLB_in_to_BLE_in = 95e-12
delay_crossbar_BLE_out_to_BLE_in = 75e-12

delay_FF_T_setup = 66e-12
delay_FF_T_clk_Q = 124e-12

delay_LUT_out_to_BLE_out = 25e-12
delay_FF_out_to_BLE_out = 45e-12

############################
##### general functions ####
############################

#Hardcode C value for interconnect for given technology
def wire_C_per_m(tech):
    if (tech == 130):
        #take from VTR's wire_C_scaling() function
        return 285e-12
    if (tech == 45):
        #take from VTR's wire_C_scaling() function
        return 250e-12
    if (tech == 22):
        #ITRS: 2.0 pF/cm (M1 metal)
        return 200e-12
    elif (tech == 7):
	#ITRS: 1.7 pF/cm
	return 170e-12
    else:
	assert(0);

def wire_C_scaling(tech):
    if (tech == 45):
        return 1.0
    elif (tech == 130):
        return (285.0/250.0)
    elif (tech == 22):
        return (200.0/250.0)
    elif (tech == 7):
        return (170.0/250.0)

def delay_LUT(size):
    if (size == 6):
        return 261e-12
    elif (size == 5):
        return 235e-12
    elif (size == 4):
        return 225e-12
    else:
        assert(0);

def xprint(s, newline=False):
    for i in range(tabs):
        f.write('  ')
    f.write(str(s))
    
    if (newline):
        f.write('\n')        

def xbegin(s):
    global tabs
    
    xprint("<" + s)
    openlist.append(s)
    tabs = tabs + 1

def xclose():
    f.write(">\n")

def xcloseend():
    global tabs
    
    f.write("/>\n")
    openlist.pop()
    tabs = tabs - 1

def xend():
    global tabs
    
    s = openlist.pop()    
    tabs = tabs - 1
    xprint("</" + s + ">\n")    

def xcopy(fsrc):
    src = open(fsrc, 'r')
    for line in src:
        xprint(line)
    src.close()
    f.write("\n")
    
def xprop(s, v):
    f.write(" " + s + "=\"" + str(v) + "\"")
    
def xport(type, name, p, e="", c=""):
    assert (type == "input" or type == "output" or type == "clock")
        
    xbegin(type)
    xprop("name", name)
    xprop("num_pins", p)
    if (e <> ""):
        xprop("equivalent", e)
        
    if (c <> ""):
        xprop("port_class", c)
    xcloseend()
    
def xcomment(s):
    f.write("<!-- ")
    f.write(s)
    f.write(" -->\n")

def outputs_by_frac_stage(frac_stage):
    if (frac_stage == 0):
        return 1
    elif (frac_stage == 1):
        return 2
    elif (frac_stage == 2 or frac_stage == 3):
        return 4

#####################
##### CACTI data ####
#####################

def memoryEnergy_sp(size, tech_nm): #this isn't important, will use post-processing anyway to compute with cacti values
	return 1.0

def memoryHeight(size):
    return max(int(math.sqrt(size/4096)), 1) #approximately assign memory height, using 4096 as being about one CLB, the y-crossing will then be adjusted to give the exact shape

##############
##### LUT ####
##############

def xLUT(LUT_size, num_LUT):
    xbegin("pb_type")
    xprop("name", "lut" + str(LUT_size))
    xprop("blif_model", ".names")
    xprop("num_pb", num_LUT)
    xprop("class", "lut")
    xclose()
    xport("input", "in", LUT_size, c="lut_in")
    xport("output", "out", 1, c="lut_out")
    
    xbegin("delay_matrix")
    xprop("type", "max")
    xprop("in_port", "lut" + str(LUT_size) + ".in")
    xprop("out_port", "lut" + str(LUT_size) + ".out")
    xclose()
    for j in range(LUT_size):
        xprint(delay_LUT(LUT_size), True)
    xend()  # delay_matrix
    
    xend()  # pb_type lut
    
##############
##### CLB ####
##############

def xCLB(k_LUT, N_BLE, I_CLB, I_BLE, fracture_level, num_FF, crossbar_str):    
    O_LUT = 2 ** fracture_level
    O_soft = O_LUT
    O_ble = O_soft
    O_CLB = N_BLE * O_ble
    I_soft = I_BLE
    
    assert((O_LUT % num_FF) == 0)
    
    xbegin("pb_type")
    xprop("name", "clb")
    xclose()
    
    if crossbar_str == "c":        
        xport("input", "I", I_CLB, "true")
    else:
        xport("input", "I", I_CLB, "false")
        
    xport("output", "O", O_CLB, "false")
    xport("clock", "clk", 1)
    
    xbegin("interconnect")
    xclose()
    
    ble_array = "ble[" + str((N_BLE - 1)) + ":0]"
    
    # Crossbar
    cb = 0
    if crossbar_str == "c":
        xbegin("complete")
        xprop("name", "crossbar")
        xprop("input", "clb.I " + ble_array + ".out")
        xprop("output", ble_array + ".in")
        xclose()    
        
        xbegin("delay_constant")
        xprop("max", delay_crossbar_CLB_in_to_BLE_in)
        xprop("in_port", "clb.I")
        xprop("out_port", ble_array + ".in")
        xcloseend()
        
        xbegin("delay_constant")
        xprop("max", delay_crossbar_BLE_out_to_BLE_in)
        xprop("in_port", ble_array + ".out")
        xprop("out_port", ble_array + ".in")
        xcloseend()
        xend()
    else:
        cb = int(crossbar_str)
         
 
        num_inputs = I_CLB + O_CLB
        
        for i_ble in range(N_BLE):
            input_idx = 0
            
            
            incr = i_ble + 1;
            # incr = 1
            
            next_start = 0
                        
            for lut_input in range(k_LUT):
                
                xbegin("mux")
                xprop("name", "crossbar-" + str(i_ble) + "-" + str(lut_input))
                
                # Inputs                           
                inputs = []
                for i_cb in range(1, cb + 1):
                       
                        
                    input_str = ""
                    if input_idx < I_CLB:
                        input_str = "clb.I[" + str(input_idx) + "]"
                    else:
                        if fracture_level == 0:
                            input_str = "ble[" + str(input_idx - I_CLB) + "].out"
                        elif fracture_level == 1:
                            input_str = "ble[" + str((input_idx - I_CLB) / 2) + "].out[" + str(input_idx % 2) + "]"
                        else:
                            assert(0)                    
                    inputs.append(input_str)
                    
                    input_idx = (input_idx + incr)
                    if input_idx >= num_inputs:
                        next_start = (next_start + 1) % incr
                        input_idx = next_start                    
       
                                    
                xprop("input", ' '.join(inputs))
                
                # Output
                output = "ble[" + str(i_ble) + "].in[" + str(lut_input) + "]"
                xprop("output", output)
                xclose()
                
                
                for input in inputs:
                    xbegin("delay_constant")
                    if (input[:3] == "clb") :
                        xprop("max", delay_crossbar_CLB_in_to_BLE_in)
                    else:
                        xprop("max", delay_crossbar_BLE_out_to_BLE_in)
                    xprop("in_port", input)
                    xprop("out_port", output)
                    xcloseend()
                
                xend()
        # End crossbar   
    xbegin("complete")
    xprop("name", "clks")
    xprop("input", "clb.clk")
    xprop("output", ble_array + ".clk")
    xcloseend()

            
    xbegin("direct")
    xprop("name", "clbouts")
    xprop("input", ble_array + ".out")
    xprop("output", "clb.O")
    xcloseend()
    
    xend()  # interconnect
    
    ### BLE ###
    xbegin("pb_type")
    xprop("name", "ble")
    xprop("num_pb", N_BLE)
    xclose()
    xport("input", "in", I_BLE)
    xport("output", "out", O_ble)
    xport("clock", "clk", 1)
    
    xbegin("interconnect")
    xclose()
    i = 0
    
    # Connection from ble to soft logic
    xbegin("direct")
    xprop("name", "direct" + str(i))
    xprop("input", "ble.in")
    xprop("output", "soft_logic.in")
    xcloseend()
    i = i + 1    
    
    # Clock connections to FFs 
    for j in range(num_FF):
        xbegin("direct")
        xprop("name", "direct" + str(i))
        xprop("input", "ble.clk")
        xprop("output", "ff[" + str(j) + ":" + str(j) + "].clk")
        xcloseend()
        i = i + 1
    
    if (fracture_level == 2):
        frac_stages = 4
    else:
        frac_stages = fracture_level + 1
    
    # Connections from Soft Logic to FFs
    soft_per_ff = O_soft / num_FF
    for j in range(num_FF):
        xbegin("mux")
        xprop("name", "mux" + str(i))
        inputs = ""
        for k in range(soft_per_ff):
            inputs = inputs + "soft_logic.out[" + str(j * soft_per_ff + k) + ":" + str(j * soft_per_ff + k) + "] "
        inputs = inputs.strip()
        xprop("input", inputs)
        xprop("output", "ff[" + str(j) + ":" + str(j) + "].D")
        xclose()
        
        if (j == 0):
            xbegin("pack_pattern");
            xprop("name", "bleF0")
            xprop("in_port", "soft_logic.out[0:0]")
            xprop("out_port", "ff[0:0].D")
            xcloseend()
        
        if (frac_stages > 1):  
            if (num_FF == 1):
                xbegin("pack_pattern");                
                xprop("name", "bleF1A")
                xprop("in_port", "soft_logic.out[0:0]")
                xprop("out_port", "ff[0:0].D")
                xcloseend()
                xbegin("pack_pattern");                
                xprop("name", "bleF1B")
                xprop("in_port", "soft_logic.out[1:1]")
                xprop("out_port", "ff[0:0].D")
                xcloseend() 
            else:
                xbegin("pack_pattern");                
                xprop("name", "bleF1" + chr(65 + j))
                xprop("in_port", "soft_logic.out[" + str(j) + ":" + str(j) + "]")
                xprop("out_port", "ff[" + str(j) + ":" + str(j) + "].D")
                xcloseend()
        xend()
        i = i + 1
    
    # Connections from Soft/FF to BLE output    
    for j in range(O_ble):
        soft_logic_out = "soft_logic.out[" + str(j) + ":" + str(j) + "]"
        FF_out = "ff[" + str(j / soft_per_ff) + ":" + str(j / soft_per_ff) + "].Q"
        BLE_out = "ble.out[" + str(j) + ":" + str(j) + "]"
        
        xbegin("mux")
        xprop("name", "mux" + str(i))
        xprop("input", soft_logic_out + " " + FF_out)
        xprop("output", BLE_out)
        xclose()
        
        xbegin("delay_constant")
        xprop("max", delay_LUT_out_to_BLE_out)
        xprop("in_port", soft_logic_out)
        xprop("out_port", BLE_out)
        xcloseend()
        
        xbegin("delay_constant")
        xprop("max", delay_FF_out_to_BLE_out)
        xprop("in_port", FF_out)
        xprop("out_port", BLE_out)
        xcloseend()
        
        xend()
        i = i + 1        
    

    
    xend()  # interconnect
    
    ### SOFT LOGIC ###
    xbegin("pb_type")
    xprop("name", "soft_logic")
    xprop("num_pb", 1)
    xclose()
    
    xport("input", "in", I_soft)
    xport("output", "out", O_soft)
    

        
    for frac_stage in reversed(range(frac_stages)):
        if (frac_stage == 3):
            # This is a special configuration with 1x(k-1) and 2x(k-2) LUTs
            special_stage = True
            num_LUT = 2
            LUT_size = k_LUT - 2
            
            LUT_size_special = k_LUT - 1
            idx_special = "[" + str(LUT_size_special - 1) + ":0]"
        else:
            special_stage = False
            
            num_LUT = 2 ** frac_stage
            LUT_size = k_LUT - frac_stage
                 
        idx = "[" + str(LUT_size - 1) + ":0]"
        xbegin("mode")
        if (special_stage):
            xprop("name", "n3" + "_lut" + str(LUT_size) + "-" + str(LUT_size_special))
        else:
            xprop("name", "n" + str(num_LUT) + "_lut" + str(LUT_size))
        xclose()
        
        xbegin("interconnect")
        xclose()
        i = 0    
        
        max_shift = I_soft - LUT_size;
        
        shift_interval = int(num_LUT)
        
        # Soft Logic to LUTs
        for j in range(num_LUT):
            if (num_LUT == 1):
                shift = 0;
            elif (special_stage):
                shift = int(float(j) / (num_LUT * 2 - 1) * max_shift)
            else:
                shift = int(float(j) / (num_LUT - 1) * max_shift)
                
            soft_idx = I_soft - 1 - shift;
            
            xbegin("direct")
            xprop("name", "direct" + str(i))        
            xprop("input", "soft_logic.in[" + str(LUT_size - 1 + shift) + ":" + str(shift) + "]")
            xprop("output", "lut" + str(LUT_size) + "[" + str(j) + ":" + str(j) + "].in" + idx)
            xcloseend()
            i = i + 1
        
        # Soft Logic to special LUT
        if (special_stage):            
            xbegin("direct")
            xprop("name", "direct" + str(i))        
            xprop("input", "soft_logic.in[" + str(I_soft - 1) + ":" + str(I_soft - LUT_size_special) + "]")
            xprop("output", "lut" + str(LUT_size_special) + "[0:0].in" + idx_special)
            xcloseend()
            i = i + 1
        
        # LUTs to soft logic
        for j in range(num_LUT):
            xbegin("direct")
            xprop("name", "direct" + str(i))
            in_str = "lut" + str(LUT_size) + "[" + str(j) + ":" + str(j) + "].out"
            out_str = "soft_logic.out[" + str(j) + ":" + str(j) + "]"
            xprop("input", in_str)
            xprop("output", out_str)
            xclose()
            
            if (outputs_by_frac_stage(frac_stage) > 1):
                char = chr(65 + j)
            else:
                char = ""
            xbegin("pack_pattern")
            xprop("name", "bleF" + str(frac_stage) + char)
            xprop("in_port", in_str)
            xprop("out_port", out_str)
            xcloseend()
            xend()
            i = i + 1
            
        # Special LUT to soft Logic
        if (special_stage):
            xbegin("direct")
            xprop("name", "direct" + str(i))
            xprop("input", "lut" + str(LUT_size_special) + "[0:0].out")
            xprop("output", "soft_logic.out[" + str(num_LUT) + ":" + str(num_LUT) + "]")
            xcloseend()
            i = i + 1     
        
        xend()  # interconnect
        
        xLUT(LUT_size, num_LUT);
        
        if (special_stage):
            xLUT(LUT_size_special, 1)
        
        xend()  # mode
        
    
    xend()  # pb_type soft_logic
    
    xbegin("pb_type")
    xprop("name", "ff")
    xprop("blif_model", ".latch")
    xprop("num_pb", num_FF)
    xprop("class", "flipflop")
    xclose()
    xport("input", "D", 1, c="D")
    xport("output", "Q", 1, c="Q")
    xport("clock", "clk", 1, c="clock")
    
    xbegin("T_setup")
    xprop("value", delay_FF_T_setup)
    xprop("port", "ff.D")
    xprop("clock", "clk")
    xcloseend()
    
    xbegin("T_clock_to_Q")
    xprop("max", delay_FF_T_clk_Q)
    xprop("port", "ff.Q")
    xprop("clock", "clk")
    xcloseend()
        
    xend()  # pb_type ff
    
    xend()  # pb_type ble
    
    xbegin("fc")
    xprop("default_in_type", "frac")
    xprop("default_in_val", Fc_in)
    xprop("default_out_type", "frac")
    xprop("default_out_val", Fc_out)
    xcloseend()
    
    xbegin("pinlocations")
    xprop("pattern", "spread")
    xcloseend()
    
    xbegin("gridlocations")
    xclose()
    xbegin("loc")
    xprop("type", "fill")
    xprop("priority", 1)
    xcloseend()
    xend()
    
    xend()  # pb_type clb

def xT_setup(port):
    xbegin("T_setup")
    xprop("value", "509e-12")
    xprop("port", port)
    xprop("clock", "clk")
    xcloseend()
    
#################
##### Mem sp ####
#################

def xmemory_sp(size, width, memory, energy):
    depth = size / width
    modeName = memory + "_" + str(depth) + "x" + str(width) + "_sp"
    addr = int(math.log(depth, 2))
    xbegin("mode")
    xprop("name", modeName)
    xclose()

    xbegin("pb_type")
    xprop("name", modeName)
    #Different top-level if large memory
    if(memory=="memory_poweropt"):
        xprop("blif_model", ".subckt single_port_ram_poweropt")
    else:
        xprop("blif_model", ".subckt single_port_ram_poweropt2")
    xprop("num_pb", 1)
    xclose()
    
    xport("input", "addr", addr)
    xport("input", "data", width)
    xport("input", "writeen", 1)
    xport("input", "readen", 1)
    xport("input", "clken", 1)
    xport("output", "out", width)
    xport("clock", "clk", 1)
    xT_setup(modeName + ".addr")
    xT_setup(modeName + ".data")
    xT_setup(modeName + ".writeen")
    xT_setup(modeName + ".readen")
    xT_setup(modeName + ".clken")

    xbegin("T_clock_to_Q")
    xprop("max", "1.234e-9")
    xprop("port", modeName + ".out")
    xprop("clock", "clk")
    xcloseend()

    xbegin("power")
    xprop("method", "pin-toggle")
    xclose()
    xbegin("port")
    xprop("name", "clk")
    xprop("energy_per_toggle", str(energy) + "e-12")
    xprop("scaled_by_static_prob", "clken")
    xcloseend()
    xbegin("static_power")
    xprop("power_per_instance", 0.0)
    xcloseend()
    xend() # power
        
    xend() # single port pb_type

    xbegin("interconnect")
    xclose()
    
    xbegin("direct")
    xprop("name", "address1")
    xprop("input", memory + ".addr1[" + str(addr - 1) + ":0]")
    xprop("output", modeName + ".addr")
    xclose()

    xbegin("delay_constant")
    xprop("max", "132e-12")
    xprop("in_port", memory + ".addr1[" + str(addr - 1) + ":0]")
    xprop("out_port", modeName + ".addr")
    xcloseend()
    
    xend() # direct

    xbegin("direct")
    xprop("name", "data1")
    xprop("input", memory + ".data[" + str(width - 1) + ":0]")
    xprop("output", modeName + ".data")
    xclose()

    xbegin("delay_constant")
    xprop("max", "132e-12")
    xprop("in_port", memory + ".data[" + str(width - 1) + ":0]")
    xprop("out_port", modeName + ".data")
    xcloseend()
    
    xend() # direct

    xbegin("direct")
    xprop("name", "wen1")
    xprop("input", memory + ".writeen1")
    xprop("output", modeName + ".writeen")
    xclose()

    xbegin("delay_constant")
    xprop("max", "132e-12")
    xprop("in_port", memory + ".writeen1")
    xprop("out_port", modeName + ".writeen")
    xcloseend()
    
    xend() # direct

    xbegin("direct")
    xprop("name", "ren1")
    xprop("input", memory + ".readen1")
    xprop("output", modeName + ".readen")
    xclose()

    xbegin("delay_constant")
    xprop("max", "132e-12")
    xprop("in_port", memory + ".readen1")
    xprop("out_port", modeName + ".readen")
    xcloseend()
    
    xend() # direct

    xbegin("direct")
    xprop("name", "en")
    xprop("input", memory + ".clken")
    xprop("output", modeName + ".clken")
    xclose()

    xbegin("delay_constant")
    xprop("max", "132e-12")
    xprop("in_port", memory + ".readen1")
    xprop("out_port", modeName + ".clken")
    xcloseend()
    
    xend() # direct

    xbegin("direct")
    xprop("name", "dataout1")
    xprop("input", modeName + ".out")
    xprop("output", memory + ".out[" + str(width - 1) + ":0]")
    xclose()

    xbegin("delay_constant")
    xprop("max", "40e-12")
    xprop("in_port", modeName + ".out")
    xprop("out_port", memory + ".out[" + str(width - 1) + ":0]")
    xcloseend()
    
    xend() # direct

    xbegin("direct")
    xprop("name", "clk")
    xprop("input", memory + ".clk")
    xprop("output", modeName + ".clk")
    xclose()

    xend() # direct
    
    xend() # interconnect
    xend() # single port mode

######################
##### mem dp true ####
######################

def xmemory_dp(size, width, memory, energy):
    depth = size / width
    modeName = memory + "_" + str(depth) + "x" + str(width) + "_dp"
    addr = int(math.log(depth, 2))
    xbegin("mode")
    xprop("name", modeName)
    xclose()

    xbegin("pb_type")
    xprop("name", modeName)
    #Different top-level if large memory
    if(memory=="memory_poweropt"):
        xprop("blif_model", ".subckt dual_port_ram_poweropt")
    else:
        xprop("blif_model", ".subckt dual_port_ram_poweropt2")
    xprop("num_pb", 1)
    xclose()
    
    xport("input", "addr1", addr)
    xport("input", "addr2", addr)
    xport("input", "data1", width)
    xport("input", "data2", width)
    xport("input", "writeen1", 1)
    xport("input", "writeen2", 1)
    xport("input", "readen1", 1)
    xport("input", "readen2", 1)
    xport("input", "clken", 1)
    xport("output", "out1", width)
    xport("output", "out2", width)
    xport("clock", "clk", 1)
    xT_setup(modeName + ".addr1")
    xT_setup(modeName + ".data1")
    xT_setup(modeName + ".writeen1")
    xT_setup(modeName + ".readen1")
    xT_setup(modeName + ".addr2")
    xT_setup(modeName + ".data2")
    xT_setup(modeName + ".writeen2")
    xT_setup(modeName + ".readen2")
    xT_setup(modeName + ".clken")

    xbegin("T_clock_to_Q")
    xprop("max", "1.234e-9")
    xprop("port", modeName + ".out1")
    xprop("clock", "clk")
    xcloseend()

    xbegin("T_clock_to_Q")
    xprop("max", "1.234e-9")
    xprop("port", modeName + ".out2")
    xprop("clock", "clk")
    xcloseend()

    xbegin("power")
    xprop("method", "pin-toggle")
    xclose()
    xbegin("port")
    xprop("name", "clk")
    xprop("energy_per_toggle", str(energy) + "e-12")
    xprop("scaled_by_static_prob", "clken")
    xcloseend()
    xbegin("static_power")
    xprop("power_per_instance", 0.0)
    xcloseend()
    xend() # power

    xend() # dual port mp_type

    xbegin("interconnect")
    xclose()
    
    xbegin("direct")
    xprop("name", "address1")
    xprop("input", memory + ".addr1[" + str(addr - 1) + ":0]")
    xprop("output", modeName + ".addr1")
    xclose()

    xbegin("delay_constant")
    xprop("max", "132e-12")
    xprop("in_port", memory + ".addr1[" + str(addr - 1) + ":0]")
    xprop("out_port", modeName + ".addr1")
    xcloseend()
    
    xend() # direct

    xbegin("direct")
    xprop("name", "address2")
    xprop("input", memory + ".addr2[" + str(addr - 1) + ":0]")
    xprop("output", modeName + ".addr2")
    xclose()

    xbegin("delay_constant")
    xprop("max", "132e-12")
    xprop("in_port", memory + ".addr2[" + str(addr - 1) + ":0]")
    xprop("out_port", modeName + ".addr2")
    xcloseend()
    
    xend() # direct

    xbegin("direct")
    xprop("name", "data1")
    xprop("input", memory + ".data[" + str(width - 1) + ":0]")
    xprop("output", modeName + ".data1")
    xclose()

    xbegin("delay_constant")
    xprop("max", "132e-12")
    xprop("in_port", memory + ".data[" + str(width - 1) + ":0]")
    xprop("out_port", modeName + ".data1")
    xcloseend()
    
    xend() # direct

    xbegin("direct")
    xprop("name", "data2")
    xprop("input", memory + ".data[" + str(2 * width - 1) + ":" + str(width) + "]")
    xprop("output", modeName + ".data2")
    xclose()

    xbegin("delay_constant")
    xprop("max", "132e-12")
    xprop("in_port", memory + ".data[" + str(2 * width - 1) + ":" + str(width) + "]")
    xprop("out_port", modeName + ".data2")
    xcloseend()
    
    xend() # direct

    xbegin("direct")
    xprop("name", "we1")
    xprop("input", memory + ".writeen1")
    xprop("output", modeName + ".writeen1")
    xclose()

    xbegin("delay_constant")
    xprop("max", "132e-12")
    xprop("in_port", memory + ".writeen1")
    xprop("out_port", modeName + ".writeen1")
    xcloseend()
    
    xend() # direct

    xbegin("direct")
    xprop("name", "we2")
    xprop("input", memory + ".writeen2")
    xprop("output", modeName + ".writeen2")
    xclose()

    xbegin("delay_constant")
    xprop("max", "132e-12")
    xprop("in_port", memory + ".writeen2")
    xprop("out_port", modeName + ".writeen2")
    xcloseend()
    
    xend() # direct

    xbegin("direct")
    xprop("name", "re1")
    xprop("input", memory + ".readen1")
    xprop("output", modeName + ".readen1")
    xclose()

    xbegin("delay_constant")
    xprop("max", "132e-12")
    xprop("in_port", memory + ".readen1")
    xprop("out_port", modeName + ".readen1")
    xcloseend()
    
    xend() # direct

    xbegin("direct")
    xprop("name", "re2")
    xprop("input", memory + ".readen2")
    xprop("output", modeName + ".readen2")
    xclose()

    xbegin("delay_constant")
    xprop("max", "132e-12")
    xprop("in_port", memory + ".readen2")
    xprop("out_port", modeName + ".readen2")
    xcloseend()
    
    xend() # direct

    xbegin("direct")
    xprop("name", "clken")
    xprop("input", memory + ".clken")
    xprop("output", modeName + ".clken")
    xclose()

    xbegin("delay_constant")
    xprop("max", "132e-12")
    xprop("in_port", memory + ".clken")
    xprop("out_port", modeName + ".clken")
    xcloseend()
    
    xend() # direct

    xbegin("direct")
    xprop("name", "dataout1")
    xprop("input", modeName + ".out1")
    xprop("output", memory + ".out[" + str(width - 1) + ":0]")
    xclose()

    xbegin("delay_constant")
    xprop("max", "40e-12")
    xprop("in_port", modeName + ".out1")
    xprop("out_port", memory + ".out[" + str(width - 1) + ":0]")
    xcloseend()
    
    xend() # direct

    xbegin("direct")
    xprop("name", "dataout2")
    xprop("input", modeName + ".out2")
    xprop("output", memory + ".out[" + str(2 * width - 1) + ":" + str(width) + "]")
    xclose()

    xbegin("delay_constant")
    xprop("max", "40e-12")
    xprop("in_port", modeName + ".out2")
    xprop("out_port", memory + ".out[" + str(2 * width - 1) + ":" + str(width) + "]")
    xcloseend()
    
    xend() # direct

    xbegin("direct")
    xprop("name", "clk")
    xprop("input", memory + ".clk")
    xprop("output", modeName + ".clk")
    xclose()

    xend() # direct
    
    xend() # interconnect
    xend() # dual port mode

########################
##### mem dp simple ####
########################

def xmemory_dp_s(size, width, memory, energy):
    depth = size / width
    modeName = memory + "_" + str(depth) + "x" + str(width) + "_dp_s"
    addr = int(math.log(depth, 2))
    xbegin("mode")
    xprop("name", modeName)
    xclose()

    xbegin("pb_type")
    xprop("name", modeName)
    #Different top-level if large memory
    if(memory=="memory_poweropt"):
        xprop("blif_model", ".subckt dual_port_ram_poweropt_s")
    else:
        xprop("blif_model", ".subckt dual_port_ram_poweropt_s2")
    xprop("num_pb", 1)
    xclose()
    
    xport("input", "addr1", addr)
    xport("input", "addr2", addr)
    xport("input", "data1", width)
    xport("input", "writeen1", 1)
    xport("input", "readen2", 1)
    xport("input", "clken", 1)
    xport("output", "out2", width)
    xport("clock", "clk", 1)
    xT_setup(modeName + ".addr1")
    xT_setup(modeName + ".data1")
    xT_setup(modeName + ".writeen1")
    xT_setup(modeName + ".clken")
    xT_setup(modeName + ".addr2")
    xT_setup(modeName + ".readen2")

    xbegin("T_clock_to_Q")
    xprop("max", "1.234e-9")
    xprop("port", modeName + ".out2")
    xprop("clock", "clk")
    xcloseend()

    xbegin("power")
    xprop("method", "pin-toggle")
    xclose()
    xbegin("port")
    xprop("name", "clk")
    xprop("energy_per_toggle", str(energy) + "e-12")
    xprop("scaled_by_static_prob", "clken")
    xcloseend()
    xbegin("static_power")
    xprop("power_per_instance", 0.0)
    xcloseend()
    xend() # power
    
    xend() # dual port s pb_type

    xbegin("interconnect")
    xclose()
    
    xbegin("direct")
    xprop("name", "address1")
    xprop("input", memory + ".addr1[" + str(addr - 1) + ":0]")
    xprop("output", modeName + ".addr1")
    xclose()

    xbegin("delay_constant")
    xprop("max", "132e-12")
    xprop("in_port", memory + ".addr1[" + str(addr - 1) + ":0]")
    xprop("out_port", modeName + ".addr1")
    xcloseend()
    
    xend() # direct

    xbegin("direct")
    xprop("name", "address2")
    xprop("input", memory + ".addr2[" + str(addr - 1) + ":0]")
    xprop("output", modeName + ".addr2")
    xclose()

    xbegin("delay_constant")
    xprop("max", "132e-12")
    xprop("in_port", memory + ".addr2[" + str(addr - 1) + ":0]")
    xprop("out_port", modeName + ".addr2")
    xcloseend()
    
    xend() # direct

    xbegin("direct")
    xprop("name", "data1")
    xprop("input", memory + ".data[" + str(width - 1) + ":0]")
    xprop("output", modeName + ".data1")
    xclose()

    xbegin("delay_constant")
    xprop("max", "132e-12")
    xprop("in_port", memory + ".data[" + str(width - 1) + ":0]")
    xprop("out_port", modeName + ".data1")
    xcloseend()
    
    xend() # direct

    xbegin("direct")
    xprop("name", "we1")
    xprop("input", memory + ".writeen1")
    xprop("output", modeName + ".writeen1")
    xclose()

    xbegin("delay_constant")
    xprop("max", "132e-12")
    xprop("in_port", memory + ".writeen1")
    xprop("out_port", modeName + ".writeen1")
    xcloseend()

    xend() # direct

    xbegin("direct")
    xprop("name", "re2")
    xprop("input", memory + ".readen2")
    xprop("output", modeName + ".readen2")
    xclose()

    xbegin("delay_constant")
    xprop("max", "132e-12")
    xprop("in_port", memory + ".readen2")
    xprop("out_port", modeName + ".readen2")
    xcloseend()
    
    xend() # direct

    xbegin("direct")
    xprop("name", "clken")
    xprop("input", memory + ".clken")
    xprop("output", modeName + ".clken")
    xclose()

    xbegin("delay_constant")
    xprop("max", "132e-12")
    xprop("in_port", memory + ".clken")
    xprop("out_port", modeName + ".clken")
    xcloseend()
    
    xend() # direct

    xbegin("direct")
    xprop("name", "dataout2")
    xprop("input", modeName + ".out2")
    xprop("output", memory + ".out[" + str(width - 1) + ":0]")
    xclose()

    xbegin("delay_constant")
    xprop("max", "40e-12")
    xprop("in_port", modeName + ".out2")
    xprop("out_port", memory + ".out[" + str(width - 1) + ":0]")
    xcloseend()
    
    xend() # direct

    xbegin("direct")
    xprop("name", "clk")
    xprop("input", memory + ".clk")
    xprop("output", modeName + ".clk")
    xclose()

    xend() # direct
    
    xend() # interconnect
    xend() # dual port s mode

#########################
##### mem interfaces ####
#########################

def xMemory(memory, size, max_width, modes, freq, sf, tech_nm):
    xbegin("pb_type")
    xprop("name", memory)
    xprop("height", memoryHeight(size))
    xclose()

    energy = memoryEnergy_sp(size, tech_nm)

    max_addr = int(math.log(size, 2))
    xport("input", "addr1", max_addr)
    xport("input", "addr2", max_addr)
    xport("input", "writeen1", 1)
    xport("input", "writeen2", 1)
    xport("input", "readen1", 1)
    xport("input", "readen2", 1)
    xport("input", "clken", 1)
    xport("input", "data", max_width)
    xport("output", "out", max_width)
    xport("clock", "clk", 1)
    xcomment("Single Ported RAMs")

    width = max_width
    for i in xrange(modes):
        xmemory_sp(size, width, memory, energy)
        width /= 2

    xcomment("Dual Ported RAMs, true dual port")

    width = max_width
    for i in xrange(modes - 1):
        width /= 2
        xmemory_dp(size, width, memory, energy * 2)
        
    xcomment("Dual Ported RAMs, simple dual port")
    width = max_width
    for i in xrange(modes):
        xmemory_dp_s(size, width, memory, energy * 2)
        width /= 2
    
    xbegin("fc")
    xprop("default_in_type", "frac")
    xprop("default_in_val", 0.15)
    xprop("default_out_type", "frac")
    xprop("default_out_val", 0.10)
    xcloseend()

    xbegin("pinlocations")
    xprop("pattern", "spread")
    xcloseend()

    if size > 2**30: #limit study
        start = 2
        freq = 2
    else:
        start = freq

    if(memory=="memory_poweropt"):
	xbegin("gridlocations")
	xclose()
	xbegin("loc")
	xprop("type", "col")
	xprop("start", start)
	xprop("repeat", freq)
	xprop("priority", 10)
	xcloseend()
	xend()
    elif(memory=="memory_poweropt2"):
	xbegin("gridlocations")
	xclose()
	xbegin("loc")
	xprop("type", "col")
	xprop("start", start*(sf+1))
	xprop("repeat", freq*(sf+1))
	xprop("priority", 20)
	xcloseend()
	xend()

    xbegin("power")
    xprop("method", "sum-of-children")
    xcloseend()

    xend() # pb_type
    

###################
##### gen_arch ####
###################

def gen_arch(dir, k_LUT, N_BLE, I_CLB, I_BLE, fracture_level, num_FF, seg_length, tech_nm, crossbar_str, mem1, mem2):
    print(dir)
    global f
    global tabs


    assert(fracture_level == 0 or fracture_level == 1 or fracture_level == 2)
    
    depth = mem1[0]
    max_width = mem1[1]
    modes = mem1[2]
    freq = mem1[3]
    size = max_width*depth
    mem_shape = str(depth) + "x" + str(max_width)
    
    if mem2 != 0:
        depth2 = mem2[0]
        max_width2 = mem2[1]
        modes2 = mem2[2]
        size2 = max_width2*depth2
        mem_shape2 = str(depth2) + "x" + str(max_width2)
    else:
        depth2 = depth
        max_width2 = max_width
        modes2 = modes
        size2 = 1
        mem_shape2 = mem_shape
    
    if crossbar_str != "c":
        filename = "k" + str(k_LUT) + "_N" + str(N_BLE) + "_I" + str(I_CLB) + "_Fi" + str(I_BLE) + "_L" + str(seg_length) + "_frac" + str(fracture_level) + "_ff" + str(num_FF) + "_C" + crossbar_str + "_" + str(tech_nm) + "nm.xml"    
    else:
        if mem2 != 0:
            filename = "k" + str(k_LUT) + "_N" + str(N_BLE) + "_L" + str(seg_length) + "_mem" + mem_shape + "_MEM" + mem_shape2 + "_poweropt2_freq" + str(freq) + "_tech" + str(tech_nm) + "nm.xml"
        else:
            if depth > 2**30:
                mem_shape = str(100000) + "x" + str(max_width)
                depth = str(123456789)
            filename = "k" + str(k_LUT) + "_N" + str(N_BLE) + "_L" + str(seg_length) + "_mem" + mem_shape + "_MEM0_poweropt2_freq" + str(freq) + "_tech" + str(tech_nm) + "nm.xml"                

    tabs = 0

    m = re.search("(.*PowerOpt)", sys.path[0])
    vtr_flow_dir = m.group(1)
    script_dir = os.path.join(vtr_flow_dir, "gen_arch")
    
    models = ["multiply", "single_port_ram_poweropt", "dual_port_ram_poweropt", "dual_port_ram_poweropt_s"]
    if mem2 != 0:
	models = ["multiply", "single_port_ram_poweropt", "dual_port_ram_poweropt", "dual_port_ram_poweropt_s", "single_port_ram_poweropt2", "dual_port_ram_poweropt2", "dual_port_ram_poweropt_s2"]

    #multiplier has different frequency depending on the memory frequency (should always start at 4th logic, then every 16th column)
    mult_36_to_use = ""
    if freq==2:
	mult_36_to_use = "mult_36_2"
    if freq==3:
	mult_36_to_use = "mult_36_3"
    if freq==4:
	mult_36_to_use = "mult_36_4"
    if freq==5:
	mult_36_to_use = "mult_36_5"
    if freq==6:
	mult_36_to_use = "mult_36_6"
    if freq==7:
	mult_36_to_use = "mult_36_7"
    if freq==8:
	mult_36_to_use = "mult_36_8"
    if freq==9:
	mult_36_to_use = "mult_36_9"
    if freq==10:
	mult_36_to_use = "mult_36_10"
    if freq==11:
	mult_36_to_use = "mult_36_11"
    if freq==12:
	mult_36_to_use = "mult_36_12"
    if freq==13:
	mult_36_to_use = "mult_36_13"
    if freq==14:
	mult_36_to_use = "mult_36_14"
    if freq==15:
	mult_36_to_use = "mult_36_15"
    if freq==16:
	mult_36_to_use = "mult_36_16"
    
    if depth > 2**30:
	cbs = ["io", "clb", "mult_36_limit", "memory"]
    else:
	cbs = ["io", "clb", mult_36_to_use, "memory"]
    
    if not os.path.exists(dir):
        os.makedirs(dir)

    f = open(os.path.join(dir, filename), 'w')
    
    xcomment("depth:_" + str(depth) + "_ width:_" + str(max_width) + "_")
    if mem2 != 0:
        xcomment("depth2:_" + str(depth2) + "_ width2:_" + str(max_width2) + "_")
    xcomment("k = " + str(k_LUT))
    xcomment("N = " + str(N_BLE))
    xcomment("I = " + str(I_CLB))
    xcomment("Fi = " + str(I_BLE))
    xcomment("frac_level = " + str(fracture_level))
    xcomment("ff = " + str(num_FF))
    
    xbegin("architecture")
    xclose()
    
    xbegin("models")
    xclose()
    
    for model in models: 
        xcopy(os.path.join(script_dir, "models", model + ".xml"))
    
    xend()  # Models
    
    xbegin("layout")
    xprop("auto", "1.0")
    xcloseend()
    xbegin("device")
    xclose()
    
    xbegin("sizing")
    xprop("R_minW_nmos", R_minW_nmos)
    xprop("R_minW_pmos", R_minW_pmos)
    xprop("ipin_mux_trans_size", ipin_mux_trans_size)
    xcloseend()
    
    xbegin("timing")
    xprop("C_ipin_cblock", C_ipin_cblock * wire_C_scaling(tech_nm))
    xprop("T_ipin_cblock", T_ipin_cblock)
    xcloseend()
    
    xbegin("area")
    xprop("grid_logic_tile_area", grid_logic_tile_area)
    xcloseend()
    
    xbegin("chan_width_distr")
    xclose()
    
    xbegin("io")
    xprop("width", 1.0)
    xcloseend()
    
    xbegin("x")
    xprop("distr", "uniform")
    xprop("peak", "1.0")
    xcloseend()
    
    xbegin("y")
    xprop("distr", "uniform")
    xprop("peak", "1.0")
    xcloseend()
    
    xend()  # chan_width_distr
    
    xbegin("switch_block")
    xprop("type", "wilton")
    xprop("fs", 3)
    xcloseend()
    
    xend()  # Device
    
    xbegin("switchlist")
    xclose()
    xbegin("switch")
    xprop("type", "mux")
    xprop("name", "0")
    xprop("R", switch_R)
    xprop("Cin", switch_Cin * wire_C_scaling(tech_nm))
    xprop("Cout", switch_Cout * wire_C_scaling(tech_nm))
    xprop("Tdel", switch_T_del)
    xprop("mux_trans_size", switch_mux_trans_size)
    xprop("buf_size", switch_buf_size)
    xcloseend()
    xend()  # switchlist
    
    xbegin("segmentlist")
    xclose()
    xbegin("segment")
    xprop("freq", 1.0)
    xprop("length", seg_length)
    xprop("type", "unidir")
    xprop("Rmetal", segment_Rmetal)
    xprop("Cmetal", segment_Cmetal * wire_C_scaling(tech_nm))
    # if (do_power):
    #    xprop("Cmetal_per_m", C_wire_glb(tech_nm))
    xclose()
    
    xbegin("mux")
    xprop("name", "0")
    xcloseend()
    
    xbegin("sb")
    xprop("type", "pattern")
    xclose()
    s = ""
    for i in range(seg_length + 1):
        s = s + "1 "
    s = s.strip()
    xprint(s + "\n", False)
    xend()  # sb
    
       
    xbegin("cb")
    xprop("type", "pattern")
    xclose()
    
    s = ""
    for i in range(seg_length):
        s = s + "1 "
    s = s.strip()
    xprint(s + "\n", False)
    xend()  # cb
    
    xend()  # segment
    xend()  # segmentlist
    
    xbegin("complexblocklist")
    xclose()

    #This loop either calls external complexblocks, or functions in this file for CLBs and memories
    for cb in cbs: 
        if (cb == "clb"):
            xCLB(k_LUT, N_BLE, I_CLB, I_BLE, fracture_level, num_FF, crossbar_str)
        elif (cb == "memory"):
	    #Compute sf: spacing factor between memories (doesn't matter for one mem size)
	    sf = 1
	    sf = int(math.sqrt(size2 / size))

            if mem2 == 0:
                xMemory("memory_poweropt", size, max_width, modes, freq, sf, tech_nm)
            else:
                xMemory("memory_poweropt", size, max_width, modes, freq, sf, tech_nm)
                xMemory("memory_poweropt2", size2, max_width2, modes2, freq, sf, tech_nm)
        else:
            xcopy(os.path.join(script_dir, "complexblocks", str(tech_nm) + "nm", cb + ".xml"))
     
    xend()  # complexblocklist
    
    if (do_power):
        xbegin("power")
        xclose()
        
        # xbegin("short_circuit_power")
        # xprop("percentage", power_short_circuit_percentage)
        # xcloseend()
        
        xbegin("local_interconnect")
        xprop("C_wire", wire_C_per_m(tech_nm))
        xcloseend()
        
        xend()  # Power
        
        xbegin("clocks")
        xclose()
        xbegin("clock")
        xprop("buffer_size", "auto")
        xprop("C_wire", wire_C_per_m(tech_nm))
        xcloseend()
        xend()  # clocks
    
    xend()  # Architecture
    
    f.close()
    print filename, "...done"


##########################
##### Sweep functions ####
##########################

# Architecture Options
arches = []

dir = ""

# K - N - I - Fi - Frac - FF - L - nm - crossbar - mem1, mem2
# mem tuple: mem_size (Kb), max_width, num_modes, freq

def mem1sweep():
    global dir
    dir = "mem1/"
    
    techs = [22]
    #widths = [16, 32, 64, 128, 256]
    #depths = [32, 64, 128, 256, 512, 1024, 2048, 4096, 8192, 16384, 32768, 65536, 131072, 262144, 524288, 1048576]
    #freq = [2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16]
    widths = [32]
    depths = [32, 64, 128, 256, 512, 1024]
    freq = [4]
    for t in techs:
        for w in widths:
            for d in depths:
                for f in freq:
		    m = int(math.log(w, 2))+1
                    mem1 = d, w, m, f
                    mem2 = 0
                    arches.append([4, 10, 22, 4, 0, 1, 1, t, "c", mem1, mem2])

def limitSweep():
    global dir
    dir = "limit/"
    mem1 = 2**59, 32, 6, 2
    mem2 = 0
    arches.append([4, 10, 22, 4, 0, 1, 1, 22, "c", mem1, mem2])

def mem2sweep():
    global dir
    dir = "mem2/"
    techs = [22]
    #widths = [16, 32, 64, 128, 256]
    #depths = [32, 64, 128, 256, 512, 1024, 2048, 4096, 8192, 16384, 32768, 65536, 131072, 262144, 524288, 1048576]
    #freq = [2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16]
    widths = [32]
    depths = [32, 64, 128, 256, 512, 1024]
    freq = [4]
    for t in techs:
        for w1 in widths:
            for w2 in widths:
                for d1 in depths:
                    for d2 in depths:
                        for f in freq:
	                    if (d2>d1 and w2>=w1 and (d2*w2)>=16*(d1*w1)):
		    		m1 = int(math.log(w1, 2))+1
		    		m2 = int(math.log(w2, 2))+1
                                mem1 = d1, w1, m1, f
                                mem2 = d2, w2, m2, f
                                arches.append([4, 10, 22, 4, 0, 1, 1, t, "c", mem1, mem2])


###############
##### Main ####
###############

# Below choose which architectures to generate (comment ones that are not needed)

mem1sweep()
#limitSweep()
#mem2sweep()

### Note: if more than one arch is to be generated, all the files will be written
### to the directory assocaited w/ the last function call !!!

for arch in arches:
    gen_arch(dir, *arch)

